#pragma ident	"@(#)sunMagellan.c	35.2	96/02/08 SMI"

/************************************************************
Copyright 1993 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or MIT not be used in
advertising or publicity pertaining to distribution  of  the
software  without specific prior written permission. Sun and
M.I.T. make no representations about the suitability of this
software for any purpose. It is provided "as is" without any
express or implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

/*|-------------------------------------------------------------------|*/
/*|   Sun X Input Extension Driver for the Magellan 3D Controller     |*/ 
/*|                    SunOS 5.5 Solaris 2.5                          |*/
/*|-------------------------------------------------------------------|*/
/*|                                                                   |*/
/*| Logitech Inc.                                           02/22/96  |*/
/*| 6505 Kaiser Drive                                                 |*/
/*| Fremont, CA 94555-3615                                            |*/
/*|                                                                   |*/
/*| Author: Volker Senft                                              |*/
/*|                                                                   |*/
/*| Tel: (510) 713 4055 (USA)                                         |*/
/*|      ++49 8153 28 2469 (Germany)                                  |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/

/*|-------------------------------------------------------------------|*/
/*| Installation hints:                                               |*/
/*|-------------------------------------------------------------------|*/
/*|                                                                   |*/
/*|  Copy the object file "ddxSUNWmagellan.so.1" in the directory     |*/
/*|  "/usr/openwin/server/modules".                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
/*|                                                                   |*/
/*|  Enter in the configuration file "OWconfig", in the directory     |*/
/*|  "/usr/openwin/server/etc" following lines.                       |*/
/*|                                                                   |*/
/*|   # sun Magellan 3D Controller module                             |*/
/*|   # Magellan 3D Controller, Logitech Inc.                         |*/
/*|   class="XINPUT" name="magellan"                                  |*/
/*|            ddxHandler="ddxSUNWmagellan.so.1"                      |*/
/*|            ddxInitFunc="ddxSUNWmagellanProc"                      |*/
/*|            ddxSerialPort="/dev/ttya";                             |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
/*|                                                                   |*/
/*| Start your application or read the documentation of the           |*/
/*| X Input Extension interface in the book:                          |*/
/*|                                                                   |*/
/*| The X Resource Issue Four                                         |*/
/*| The X11 Input Extension: Tutorial and Documentation               |*/
/*|                                                                   |*/
/*| ISBN 0-937175-99-4                                                |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/

#define NEED_EVENTS

#include <stdio.h>
#include <fcntl.h>
#include <termio.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <poll.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/termiox.h>
#include <sys/filio.h>
#include <sys/stropts.h>
#include <sys/systm.h>

#include <Sunowconfig.h>

#include "XI.h"
#include "os.h"
#include "sun.h"
#include "sunInput.h"

/*|-------------------------------------------------------------------|*/
/*| MAGELLAN_DEBUG                                                    |*/
/*|                                                                   |*/
/*| Define this symbol to enable the debug output.                    |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
/*| #define MAGELLAN_DEBUG                                            |*/
/*|-------------------------------------------------------------------|*/
#ifdef MAGELLAN_DEBUG 
#define MagellanDebugF(FORMAT,PARAMETER)	ErrorF(FORMAT,PARAMETER)
#define MagellanDebug(STRING)			Error(STRING)
#else
#define MagellanDebugF(FORMAT,PARAMETER)
#define MagellanDebug(STRING)
#endif
/*|-------------------------------------------------------------------|*/


/*|-------------------------------------------------------------------|*/
/*| Macros                                                            |*/ 
/*|                                                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
#ifndef TRUE
#define TRUE	(1==1)
#endif

#ifndef FALSE
#define FALSE	(1==0)
#endif

#define MagellanNibble(Value)		(Value&0x0F)
#define MAGELLAN_RESOLUTION		200
#define MAGELLAN_MIN			-16384
#define MAGELLAN_MAX			+16384

enum _MagellanButtons_ { MagellanButton1, MagellanButton2, MagellanButton3, MagellanButton4,
                         MagellanButton5, MagellanButton6, MagellanButton7, MagellanButton8, MagellanButtonPick,
                         MagellanButtons };

enum _MagellanValuators_ { X,Y,Z, A,B,C, MagellanValuators };

enum _MagellanEvents_ { MagellanMotionEvent=27695, MagellanButtonPressEvent, MagellanButtonReleaseEvent };

/*|-------------------------------------------------------------------|*/
/*| Device Private Structures                                         |*/ 
/*|                                                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
struct _sunMagellanPrivate_ {
 int handle;		/*| File descriptor of the Magellan 3D Controller |*/
 Atom atom;		/*| Atom of the Magellan 3D Controller, Type of device |*/
};
typedef struct _sunMagellanPrivate_ sunMagellanPrivate;

union _MagellanXEvent_ {
 deviceValuator EventMotion;
 deviceKeyButtonPointer EventButton;
};
typedef union _MagellanXEvent_ MagellanXEvent;

struct _MagellanEvent_ {
 int EventType;
 int EventButton;
 MagellanXEvent EventX;
};
typedef struct _MagellanEvent_ MagellanEvent;

/*|-------------------------------------------------------------------|*/
/*| Function Prototypes                                               |*/ 
/*|                                                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
int 	ddxSUNWmagellanProc(DevicePtr, int);
XI_eventPtr sunMagellanGetEvents(DevicePtr, int *, Bool *);
void 	sunMagellanEnqueueEvent(DevicePtr, XI_eventPtr);
int 	sunMagellanMotionProc(DeviceIntPtr, INT32 *, unsigned long, unsigned long, ScreenPtr);

/*|-------------------------------------------------------------------|*/
/*| External Declarations                                             |*/ 
/*|                                                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
extern int DeviceMotionNotify;
extern int DeviceButtonPress;
extern int DeviceButtonRelease;
extern int DeviceValuator;

/*|-------------------------------------------------------------------|*/
/*| Constant and Variable Definitions                                 |*/ 
/*|                                                                   |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
#define MagellanBufferSize	256
static char MagellanBuffer[ MagellanBufferSize ];
static char MagellanString[ MagellanBufferSize ];
static char MagellanSerialPortBuffer[ MagellanBufferSize ]; 
static int  MagellanIndex = 0;

static char MagellanInitString[] = "\rz\rz\r";		/* switch in 3D mode */
static char MagellanShortBeep[] = "b9\r";		/* a short beep */
static char MagellanMode[] = "m3\r";			/* translation and rotation data ON */
static char MagellanPeriod[] = "pAA\r";			/* transmit every 60 ms data */
static char MagellanNullRadius[] = "nH\r";		/* null radius to value 8 */
static char MagellanZero[] = "z\r";			/* detect zero position */
static char MagellanSensitivity[] = "q00\r";		/* no extra sensitivity */
static char MagellanVersion[] = "vQ\r";			/* get version string */

static int MagellanKeyboard = 0;

static MagellanEvent MagellanEvents[MAXEVENTS];
static XI_event MagellanXIEvents[MAXEVENTS];

static CARD8 MagellanButtonMap[] = 
 { 0, MagellanButton1+1, MagellanButton2+1, MagellanButton3+1, 
      MagellanButton4+1, MagellanButton5+1, MagellanButton6+1, 
      MagellanButton7+1, MagellanButton8+1, MagellanButtonPick+1 };
/*|-------------------------------------------------------------------|*/



/*|-------------------------------------------------------------------|*/
/*| MagellanSerialPortName                                            |*/
/*|                                                                   |*/
/*| Determine the name of the serial port. Try to read in the         |*/
/*| OWconfig file the attribute "ddxSerialPort". This attribute       |*/ 
/*| determine the serial port. The default is "/dev/ttya".            |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
char *MagellanSerialPortName( DeviceIntPtr pMagellan )
{
char *SerialPort;

 strcpy( MagellanSerialPortBuffer, "/dev/ttya" );

 MagellanDebugF("Magellan: Serial Port Name for Device = %s \n", pMagellan->devEntry->tag );
 SerialPort = OWconfigGetAttribute( "XINPUT", pMagellan->devEntry->tag, "ddxSerialPort" );
 if ( SerialPort != NULL )
  {
   MagellanDebugF("Magellan: Serial Port Name = %s \n", SerialPort );
   strcpy( MagellanSerialPortBuffer, SerialPort );
   OWconfigFreeAttribute( SerialPort );
  }; 
 return MagellanSerialPortBuffer;
}

/*|-------------------------------------------------------------------|*/
/*| MagellanWait                                                      |*/
/*|                                                                   |*/
/*| Wait a time in Millisecond.                                       |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static void MagellanWait( WaitTime )
clock_t WaitTime;
{
clock_t start,stop;

 if ( WaitTime >= 1000 )
  {
   sleep( 1 );
   MagellanWait( WaitTime - 1000 );
  }
  else
  {
   start = clock();
   do
    {
     stop = clock();
    }while( labs(stop-start) < WaitTime*CLOCKS_PER_SEC/1000 );
  };
}

/*|-------------------------------------------------------------------|*/
/*| MagellanReadString                                                |*/
/*|                                                                   |*/
/*| Read a string from Magellan 3D Controller.                        |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static int MagellanReadString( MagellanHandle, MagellanString ) 
int MagellanHandle; 
char *MagellanString; 
{ 
char *cptr;
int result;

 do
  {
   cptr = strchr( MagellanBuffer, '\r' );		/* check if end of packet */
   if ( cptr == NULL )
    {
     result = read( MagellanHandle, &MagellanBuffer[ MagellanIndex ], /* read from the serial port */
                              sizeof(MagellanBuffer)-MagellanIndex );
     if ( result == 0 ) 
      return FALSE;    
     MagellanIndex += result;               
     MagellanBuffer[ MagellanIndex ] = '\000';
    }
   else
    {
     *cptr = '\000';
     strcpy( MagellanString, MagellanBuffer );
     strcpy( MagellanBuffer, ++cptr );
     MagellanIndex = strlen( MagellanBuffer );

     MagellanDebugF("Magellan: Read String \n%s\n", MagellanString );
     return TRUE;
    };
  } while( TRUE );
} 
 
 
/*|-------------------------------------------------------------------|*/
/*| MagellanWriteString                                               |*/
/*|                                                                   |*/
/*| Write a string to the Magellan 3D Controller.                     |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static int MagellanWriteString( MagellanHandle, MagellanString ) 
int MagellanHandle; 
char * MagellanString; 
{
 if ( MagellanHandle == NULL )
  return FALSE;
 else
  {
   MagellanDebugF("Magellan: Write String \n%s\n", MagellanString );
   write( MagellanHandle, MagellanString, strlen(MagellanString) );  
  };
 return TRUE;
} 

/*|-------------------------------------------------------------------|*/
/*| MagellanOpen                                                      |*/
/*|                                                                   |*/
/*| Open the serial port to the Magellan 3D Controller and initilize  |*/
/*| the device.                                                       |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static int MagellanOpen ( char *MagellanSerialPort ) 
{ 
int result, length, MagellanHandle; 
struct termio MagellanTermio; 
 
 MagellanHandle = open ( MagellanSerialPort, O_RDWR | O_NDELAY ); 
 if ( MagellanHandle < 0)  
  return 0;				/* Can't open serial device */ 

 MagellanDebugF("Magellan: Serial Port Handle = %d \n", MagellanHandle );
 
 /* configure the serial port for the Magellan 3D Controller */ 
 ioctl ( MagellanHandle, TCGETA, &MagellanTermio ); 
 MagellanTermio.c_lflag = 0; 
 MagellanTermio.c_iflag = IGNBRK | IGNPAR | IXON | IXOFF; 
 MagellanTermio.c_oflag = 0;         	
 MagellanTermio.c_cflag =  B9600 | CREAD | CS8 | CSTOPB | CLOCAL;
 MagellanTermio.c_cc[VEOL] = '\r';  			 
 MagellanTermio.c_cc[VERASE] = CNUL;
 MagellanTermio.c_cc[VKILL] = CNUL;
 MagellanTermio.c_cc[VMIN] = 0; 
 MagellanTermio.c_cc[VTIME] = 0; 
 ioctl ( MagellanHandle, TCSETA, &MagellanTermio ); 

 MagellanWait( 2000 );					/* wait for the Magellan power up */ 
 
 memset( MagellanBuffer, 0x00, sizeof(MagellanBuffer) );
 MagellanIndex = 0;

 MagellanWriteString( MagellanHandle, MagellanInitString );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanVersion );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanSensitivity );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanZero );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanPeriod );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanMode );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanNullRadius );
 MagellanWait( 100 );

 MagellanWriteString( MagellanHandle, MagellanShortBeep );
 
 MagellanWait( 1000 );					/* wait for request from the Magellan */

 while ( MagellanReadString( MagellanHandle, MagellanString ) )
  {
   if ( (strstr( MagellanString, "SPACE MOUSE" ) != NULL) || 
        (strstr( MagellanString, "MAGELLAN" ) != NULL) )
    {
     return MagellanHandle;
    };
  };

 if ( MagellanHandle > 0 )
  {
   close( MagellanHandle );
   MagellanHandle = 0;
  }; 

 return MagellanHandle;   
} 
 

/*|-------------------------------------------------------------------|*/
/*| MagellanRead                                                      |*/
/*|                                                                   |*/
/*| Readout from the serial port the messages of the Magellan 3D      |*/
/*| Controller and convert the 3D data and the keyboard messages to   |*/
/*| X Input Extension Events.                                         |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static int MagellanRead ( DevicePtr pMagellan, MagellanEvent *MagellanEvents, XI_event *MagellanXIEvents )
{
sunMagellanPrivate *pPrivate;
int MagellanButton, loop, mask, event;
int MagellanData, MagellanHandle;

 pPrivate = (sunMagellanPrivate *)(pMagellan->devicePrivate);
 MagellanHandle = pPrivate->handle;

 MagellanDebugF("Magellan: Enter MagellanRead ... File Handle = %d \n", MagellanHandle );

 event = 0; 
 MagellanXIEvents[event].event = (pointer) NULL;
 do 
  {
   if ( MagellanReadString( MagellanHandle, MagellanString ) == FALSE )
    {
     return event;
    }
   else
    {
     MagellanDebugF("Magellan: Read from Magellan \n%s\n", MagellanString );

     switch( MagellanString[ 0 ] )
      {
       case 'd' :
        MagellanEvents[event].EventType = MagellanMotionEvent;
        MagellanEvents[event].EventX.EventMotion.type = DeviceValuator;
        MagellanEvents[event].EventX.EventMotion.deviceid = 0;
        MagellanEvents[event].EventX.EventMotion.sequenceNumber = 0;
        MagellanEvents[event].EventX.EventMotion.device_state = 0;
        MagellanEvents[event].EventX.EventMotion.num_valuators = MagellanValuators;
        MagellanEvents[event].EventX.EventMotion.first_valuator = X;

        MagellanEvents[event].EventX.EventMotion.valuator0 =
         MagellanNibble( MagellanString[1] ) * 4096 +
         MagellanNibble( MagellanString[2] ) * 256 +
         MagellanNibble( MagellanString[3] ) * 16 +
         MagellanNibble( MagellanString[4] ) - 32768;
        MagellanEvents[event].EventX.EventMotion.valuator1 =
         MagellanNibble( MagellanString[5] ) * 4096 +
         MagellanNibble( MagellanString[6] ) * 256 +
         MagellanNibble( MagellanString[7] ) * 16 +
         MagellanNibble( MagellanString[8] ) - 32768;
        MagellanEvents[event].EventX.EventMotion.valuator2 =
         MagellanNibble( MagellanString[9] ) * 4096 +
         MagellanNibble( MagellanString[10] ) * 256 +
         MagellanNibble( MagellanString[11] ) * 16 +
         MagellanNibble( MagellanString[12] ) - 32768;

        MagellanEvents[event].EventX.EventMotion.valuator3 =
         MagellanNibble( MagellanString[13] ) * 4096 +
         MagellanNibble( MagellanString[14] ) * 256 +
         MagellanNibble( MagellanString[15] ) * 16 +
         MagellanNibble( MagellanString[16] ) - 32768;
        MagellanEvents[event].EventX.EventMotion.valuator4 =
         MagellanNibble( MagellanString[17] ) * 4096 +
         MagellanNibble( MagellanString[18] ) * 256 +
         MagellanNibble( MagellanString[19] ) * 16 +
         MagellanNibble( MagellanString[20] ) - 32768;
        MagellanEvents[event].EventX.EventMotion.valuator5 =
         MagellanNibble( MagellanString[21] ) * 4096 +
         MagellanNibble( MagellanString[22] ) * 256 +
         MagellanNibble( MagellanString[23] ) * 16 +
         MagellanNibble( MagellanString[24] ) - 32768;

        MagellanXIEvents[event].event = (pointer) &MagellanEvents[event]; 
        MagellanXIEvents[event].time.tv_sec = 0;
        MagellanXIEvents[event].time.tv_usec = 0;

        MagellanDebugF("Magellan: Create MotionEvent [%d] \n", event );

        if ( ++event >= MAXEVENTS )
         return event;
        else
         MagellanXIEvents[event].event = (pointer) NULL;
        break;
          
       case 'k' :
        MagellanButton = MagellanNibble( MagellanString[1] ) * 1 +
                         MagellanNibble( MagellanString[2] ) * 16 +
                         MagellanNibble( MagellanString[3] ) * 256;  
        for ( loop=0; loop<9; ++loop )
         {
          mask = 1 << loop;
          if ( (MagellanButton ^ MagellanKeyboard) & mask )
           {
            if ( MagellanButton & mask )
             {
              MagellanEvents[event].EventType = MagellanButtonPressEvent; 
              MagellanEvents[event].EventX.EventButton.type = DeviceButtonPress;
             }
            else
             {
              MagellanEvents[event].EventType = MagellanButtonReleaseEvent; 
              MagellanEvents[event].EventX.EventButton.type = DeviceButtonRelease;
             };
            MagellanEvents[event].EventX.EventButton.detail = loop+1;
            MagellanEvents[event].EventX.EventButton.sequenceNumber = 0;
            MagellanEvents[event].EventX.EventButton.time = 0;
            MagellanEvents[event].EventX.EventButton.root = 0;
            MagellanEvents[event].EventX.EventButton.event = 0;
            MagellanEvents[event].EventX.EventButton.child = 0;
            MagellanEvents[event].EventX.EventButton.root_x = 0;
            MagellanEvents[event].EventX.EventButton.root_y = 0;
            MagellanEvents[event].EventX.EventButton.event_x = 0;
            MagellanEvents[event].EventX.EventButton.event_y = 0;
            MagellanEvents[event].EventX.EventButton.state = 0;
            MagellanEvents[event].EventX.EventButton.same_screen = TRUE;

            MagellanXIEvents[event].event = (pointer) &MagellanEvents[event]; 
            MagellanXIEvents[event].time.tv_sec = 0;
            MagellanXIEvents[event].time.tv_usec = 0;

            MagellanDebugF("Magellan: Create ButtonEvent [%d] \n", event );

            if ( ++event >= MAXEVENTS )
             return event;
            else
             MagellanXIEvents[event].event = (pointer) NULL;
           };
         };
        MagellanKeyboard = MagellanButton;
        break;
      };
    };
  } while ( TRUE );
} 


/*|-------------------------------------------------------------------|*/
/*| MagellanKbdCtrl                                                   |*/
/*|                                                                   |*/
/*| Control the Bell. Empty function.                                 |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
void MagellanKbdCtrl( DeviceIntPtr pPtr, KeybdCtrl *ctrl )
{
 MagellanDebug("Magellan: Enter MagellanKbdCtrl ... \n");

 MagellanDebugF("Magellan: DeviceIntPtr = %08X \n", pPtr );
 MagellanDebugF("Magellan: Control = %08X \n", ctrl );
}

/*|-------------------------------------------------------------------|*/
/*| MagellanBellCtrl                                                  |*/
/*|                                                                   |*/
/*| Control the Bell. Empty function.                                 |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
void MagellanBellCtrl( DeviceIntPtr pPtr, BellCtrl *ctrl )
{
 MagellanDebug("Magellan: Enter MagellanBellCtrl ... \n");

 MagellanDebugF("Magellan: DeviceIntPtr = %08X \n", pPtr );
 MagellanDebugF("Magellan: Control = %08X \n", ctrl );
}

/*|-------------------------------------------------------------------|*/
/*| MagellanBellProc                                                  |*/
/*|                                                                   |*/
/*| Ring the Bell.                                                    |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
void MagellanBellProc( int percent, DeviceIntPtr pPtr )
{
sunMagellanPrivate *pPrivate;
int MagellanHandle;

 pPrivate = (sunMagellanPrivate *)((DevicePtr)pPtr)->devicePrivate;
 MagellanHandle = pPrivate->handle;

 MagellanDebug("Magellan: Enter MagellanBellProc ... \n");

 MagellanDebugF("Magellan: DeviceIntPtr = %08X \n", pPtr );
 MagellanDebugF("Magellan: Percent = %08X \n", percent );
 MagellanDebugF("Magellan: Handle = %d \n", MagellanHandle );

 MagellanWriteString( MagellanHandle, MagellanShortBeep );
 MagellanWait( 100 );
}

/*|-------------------------------------------------------------------|*/
/*| ddxSUNWmagellanProc                                               |*/
/*|                                                                   |*/
/*| Entry procedure for this driver. It handles the device functions  |*/
/*| INIT, OFF, ON and CLOSE.                                          |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
int ddxSUNWmagellanProc (DevicePtr pMagellan, int cmd)
{
sunMagellanPrivate *pPrivate;
int result, loop, handle;

 MagellanDebug("\n\nMagellan: Magellan 3D Controller Sun X Input Extension \n\n\n\n"); 
 MagellanDebugF("Magellan: sunMagellanProc DevicePtr = %08X\n", pMagellan );
 MagellanDebugF("Magellan: sunMagellanProc Cmd = %s \n", 
                    cmd == DEVICE_INIT  ? "DEVICE_INIT" :
                    cmd == DEVICE_OFF   ? "DEVICE_OFF" :
                    cmd == DEVICE_ON    ? "DEVICE_ON" :
                    cmd == DEVICE_CLOSE ? "DEVICE_CLOSE" : "UNKOWN" );

 switch ( cmd ) 
  {
   case DEVICE_INIT:
    if ( pMagellan->devicePrivate != NULL )
     return !Success;

    MagellanDebugF("Magellan: Serial Port Name = %s\n",MagellanSerialPortName((DeviceIntPtr) pMagellan ));
    handle = MagellanOpen( MagellanSerialPortName( (DeviceIntPtr) pMagellan ) );
    MagellanDebugF("Magellan: Serial Handle = %d\n", handle );
    if ( handle == NULL )
     return !Success;

    pPrivate = (sunMagellanPrivate *) xalloc( sizeof(sunMagellanPrivate) );
    if ( pPrivate == NULL )
     return !Success;

    pPrivate->handle = handle;

    pPrivate->atom = MakeAtom( XI_SPACEBALL, strlen(XI_SPACEBALL), FALSE );
    AssignTypeAndName( (DeviceIntPtr) pMagellan, pPrivate->atom, "MAGELLAN" );

    pMagellan->devicePrivate = (pointer) pPrivate;
    pMagellan->on = FALSE;

    result = RegisterFdIo( pMagellan, pPrivate->handle, sunMagellanGetEvents, sunMagellanEnqueueEvent );
    if ( result != Success )
     {
      xfree( pPrivate );
      return !Success;
     };
    MagellanDebug("Magellan: Register sunMagellanGetEvents and sunMagellanEnqueueEvent Procedure \n");

    result = InitButtonClassDeviceStruct( (DeviceIntPtr) pMagellan, MagellanButtons, MagellanButtonMap );
    MagellanDebugF("Magellan: Init Button Class = %d \n", result );

    result = InitValuatorClassDeviceStruct( (DeviceIntPtr) pMagellan, MagellanValuators, sunMagellanMotionProc, 0, Relative );
    MagellanDebugF("Magellan: Init Valuator Class = %d \n", result );

    for ( loop=X;  loop<=C;  ++loop )
     InitValuatorAxisStruct( (DeviceIntPtr) pMagellan, loop, MAGELLAN_MIN,MAGELLAN_MAX, MAGELLAN_RESOLUTION );

    InitFocusClassDeviceStruct( (DeviceIntPtr) pMagellan );

    InitKbdFeedbackClassDeviceStruct( (DeviceIntPtr) pMagellan, MagellanBellProc, MagellanKbdCtrl );
    InitBellFeedbackClassDeviceStruct( (DeviceIntPtr) pMagellan, MagellanBellCtrl );

    MagellanDebugF("Magellan: MagellanBellProc = %08X \n", MagellanBellProc );
    MagellanDebugF("Magellan: MagellanKbdCtrl  = %08X \n", MagellanKbdCtrl  );
    MagellanDebugF("Magellan: MagellanBellCtrl = %08X \n", MagellanBellCtrl );

    MagellanDebugF("Magellan: BellProc() = %08X \n", ((DeviceIntPtr) pMagellan)->bell->BellProc );
    MagellanDebugF("Magellan: CtrlProc() = %08X \n", ((DeviceIntPtr) pMagellan)->bell->CtrlProc );

    MagellanDebugF("Magellan: KbdBellProc() = %08X \n", ((DeviceIntPtr) pMagellan)->kbdfeed->BellProc );
    MagellanDebugF("Magellan: KbdCtrlProc() = %08X \n", ((DeviceIntPtr) pMagellan)->kbdfeed->CtrlProc );

/*| Something is wrong with the mapping of BellProc and BellCtrl:

           |	BellProc   |	CtrlProc    
    --------------------------------------
    Bell   |	BellCtrl   |	KbdCtrl
    --------------------------------------
    Kbd	   |	BellProc   |	KbdCtrl


    Change the mapping of BellProc and BellCtrl to:

           |	BellProc   |	CtrlProc    
    --------------------------------------
    Bell   |	BellProc   |	BellCtrl
    --------------------------------------
    Kbd	   |	BellProc   |	KbdCtrl				|*/

    if ( ((DeviceIntPtr) pMagellan)->bell != NULL )
     {
      ((DeviceIntPtr) pMagellan)->bell->BellProc = MagellanBellProc;
      ((DeviceIntPtr) pMagellan)->bell->CtrlProc = MagellanBellCtrl;
     };

    break;


   case DEVICE_ON:
    pPrivate = (sunMagellanPrivate *)(pMagellan->devicePrivate);

    AddEnabledDevice ( pPrivate->handle );
    pMagellan->on = TRUE;
    ioctl( pPrivate->handle, I_SETSIG, S_INPUT );
    break;


   case DEVICE_OFF:
    pPrivate = (sunMagellanPrivate *)(pMagellan->devicePrivate);

    ioctl(pPrivate->handle, I_SETSIG, NULL);
    RemoveEnabledDevice ( pPrivate->handle );
    pMagellan->on = FALSE;
    break;


   case DEVICE_CLOSE:
    pPrivate = (sunMagellanPrivate *)(pMagellan->devicePrivate);
    
    if ( pMagellan->on )
     {
      RemoveEnabledDevice( pPrivate->handle );
      pMagellan->on = FALSE;
      close( pPrivate->handle );
      xfree( pMagellan->devicePrivate );
     };
    break;
   };

 return (Success);
}
	  

/*|-------------------------------------------------------------------|*/
/*| sunMagellanEnqueueEvent                                           |*/
/*|                                                                   |*/
/*| Convert the Magellan 3D Controller Events to X Input Extension    |*/
/*| Events and put the Events in the X11 queue.                       |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static void sunMagellanEnqueueEvent (DevicePtr pMagellan, XI_eventPtr Xev)
{
sunMagellanPrivate *pPrivate;
MagellanEvent *MagellanSingleEvent;
deviceKeyButtonPointer MagellanSingleEventInit;
int loop, x,y;

 MagellanDebug("Magellan: Enter sunMagellanEnqueueEvent ... \n");

 pPrivate = (sunMagellanPrivate *)pMagellan->devicePrivate;

 for ( loop=0; loop<MAXEVENTS; ++loop )
  {
   if ( Xev[loop].event == NULL )
    return;
   else
    {
     MagellanSingleEvent = (MagellanEvent *) Xev[loop].event;
     switch( MagellanSingleEvent->EventType )
      {
       case MagellanMotionEvent:
        miPointerPosition( &x, &y );
        MagellanSingleEventInit.root_x = x;
        MagellanSingleEventInit.root_y = y;
        MagellanSingleEventInit.type = DeviceMotionNotify;
        MagellanSingleEventInit.deviceid = ((DeviceIntPtr) pMagellan)->id | MORE_EVENTS;
        MagellanSingleEventInit.detail = 0;
        MagellanSingleEventInit.time = 0;
 
        MagellanSingleEvent->EventX.EventMotion.deviceid = ((DeviceIntPtr) pMagellan)->id;
 
        MagellanDebugF("Magellan: Device ID = %d \n", MagellanSingleEvent->EventX.EventMotion.deviceid );
        MagellanDebugF("Magellan: Type = %d \n", MagellanSingleEvent->EventX.EventMotion.type );
 
        mieqEnqueue( (xEvent *)&MagellanSingleEventInit );
        mieqEnqueue( (xEvent *)&MagellanSingleEvent->EventX.EventMotion );
        break;
 
       case MagellanButtonPressEvent:
        miPointerPosition( &x, &y );
        MagellanSingleEvent->EventX.EventButton.root_x = x;
        MagellanSingleEvent->EventX.EventButton.root_y = y;
        MagellanSingleEvent->EventX.EventButton.deviceid = ((DeviceIntPtr) pMagellan)->id;
 
        MagellanDebug("Magellan: Button Press Event ... \n");
        MagellanDebugF("Magellan: Device ID = %d \n", MagellanSingleEvent->EventX.EventButton.deviceid );
        MagellanDebugF("Magellan: Type = %d \n", MagellanSingleEvent->EventX.EventButton.type );
        MagellanDebugF("Magellan: Button = %d \n", MagellanSingleEvent->EventX.EventButton.detail );
 
        mieqEnqueue( (xEvent *)&MagellanSingleEvent->EventX.EventButton );
        break;
 
       case MagellanButtonReleaseEvent:
        miPointerPosition( &x, &y );
        MagellanSingleEvent->EventX.EventButton.root_x = x;
        MagellanSingleEvent->EventX.EventButton.root_y = y;
        MagellanSingleEvent->EventX.EventButton.deviceid = ((DeviceIntPtr) pMagellan)->id;
 
        MagellanDebug("Magellan: Button Release Event ... \n");
        MagellanDebugF("Magellan: Device ID = %d \n", MagellanSingleEvent->EventX.EventButton.deviceid );
        MagellanDebugF("Magellan: Type = %d \n", MagellanSingleEvent->EventX.EventButton.type );
        MagellanDebugF("Magellan: Button = %d \n", MagellanSingleEvent->EventX.EventButton.detail );
 
        mieqEnqueue( (xEvent *)&MagellanSingleEvent->EventX.EventButton );
        break;
 
       default: 
        MagellanDebug("Magellan: Illegal Event Type ... \n");
        break;
      }; 
    };
  };
 return;
}

/*|-------------------------------------------------------------------|*/
/*| sunMagellanMotionProc                                             |*/
/*|                                                                   |*/
/*| No history is kept, just return.                                  |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static int sunMagellanMotionProc(DeviceIntPtr pPtr, INT32 *coords, unsigned long start,
	                   	    unsigned long stop, ScreenPtr pScreen)
{
 MagellanDebug("Magellan: Enter sunMagellanMotionProc ... \n");
 return 0;
}

/*|-------------------------------------------------------------------|*/
/*| sunMagellanGetEvent                                               |*/
/*|                                                                   |*/
/*| Get all Events from the Magellan 3D Controller and store the      |*/
/*| Events in a data structure for the procedure                      |*/
/*| sunMagellanEnqueueEvent                                           |*/
/*|                                                                   |*/
/*|-------------------------------------------------------------------|*/
static XI_eventPtr sunMagellanGetEvents (DevicePtr pMagellan, int *pNumEvents, Bool *pAgain)
{
 MagellanDebug("Magellan: Enter sunMagellanGetEvents ... \n");

 *pNumEvents = MagellanRead( pMagellan, MagellanEvents, MagellanXIEvents );
 *pAgain = *pNumEvents == MAXEVENTS;

 MagellanDebugF("Magellan: Create now %d event(s) \n", *pNumEvents );

 if ( *pNumEvents == 0 )
  return NULL;
 else
  return MagellanXIEvents;
}

