/*|--------------------------------------------------------------------------

 MAGELLAN X-Window XIE Library                       Version 4.00 15.10.97

 Logitech Inc.
 6505 Kaiser Drive
 Fremont, CA 94555-3615

 Volker Senft:

 Email: magellan@logitech.com
        magellan@dlr.de

----------------------------------------------------------------------------|*/
#include <string.h>
#include <math.h>
#include <stdio.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/extensions/XI.h>
#include <X11/extensions/XInput.h>

#define MagellanDEBUG
#include "xdrvlib.h"
#include "xxielib.h"




#ifndef XXIELIB_H
#define XXIELIB_H
/*|--------------------------------------------------------------------------

 MAGELLAN X-Window XIE Header                        Version 4.00 15.10.97

 Logitech Inc.
 6505 Kaiser Drive
 Fremont, CA 94555-3615

 Volker Senft:

 Email: magellan@logitech.com
        magellan@dlr.de

----------------------------------------------------------------------------|*/

int MagellanXIEInit(
#ifdef ParameterCheck
Display *display,
Window window
#endif
);

int MagellanXIEInputEvent(
#ifdef ParameterCheck
Display *display,
XEvent *Event,
MagellanIntEvent *MagellanEvent
#endif
);

int MagellanXIETranslateEvent(
#ifdef ParameterCheck
Display *display,
XEvent *Event,
MagellanFloatEvent *MagellanEvent,
double MagellanTranslationScale,
double MagellanRotationScale
#endif
);

int MagellanXIEClose(
#ifdef ParameterCheck
Display *display
#endif
);

Bool MagellanXIECheckMotionEvent(
#ifdef ParameterCheck
Display *display,
XEvent *Event,
char *Arg
#endif
);

int MagellanXIERemoveMotionEvents(
#ifdef ParameterCheck
Display *display
#endif
);

#endif  /* Magellan X Window XIE Header */





static int MagellanNotifyType[3] = { 0,0,0 };
static XDevice *MagellanXIEDevice = NULL;


int MagellanXIEInit( display, window )
Display *display;
Window window; 
{
int MagellanXIEId, DeviceNumbers, loop, MagellanExist;
XDeviceInfoPtr DeviceList;
char *DeviceTypeName;
XExtensionVersion *ExtVersion;
XEventClass MagellanNotifyClass[3];
int MagellanXIEMotionEventType;
int MagellanXIEButtonPressEventType;
int MagellanXIEButtonReleaseEventType;
int MagellanXIEMotionEventClass;
int MagellanXIEButtonPressEventClass;
int MagellanXIEButtonReleaseEventClass;

 MagellanExist = FALSE;

 ExtVersion = XGetExtensionVersion(display,"XInputExtension");
 MagellanExist = (ExtVersion != NULL) &&
                 (((int) ExtVersion) != NoSuchExtension);
 XFree( ExtVersion );
 if ( MagellanExist == FALSE )
  return FALSE;

 MagellanXIEId = FALSE;
 DeviceList = XListInputDevices(display, &DeviceNumbers);
 if( DeviceList != NULL )
  {
   for( loop=0; loop<DeviceNumbers; ++loop )
    {
     DeviceTypeName = XGetAtomName( display, DeviceList[loop].type );
#ifdef MagellanDEBUG
     printf("Device Name = %-16s  Type = %s \n",
                          DeviceList[loop].name, DeviceTypeName );
#endif
     if ( (strcmp( DeviceList[loop].name, "magellan" ) == 0) ||
          (strcmp( DeviceList[loop].name, "MAGELLAN" ) == 0) ||
          (strcmp( DeviceTypeName, "space mouse" ) == 0) ||
          (strcmp( DeviceTypeName, "SPACE MOUSE" ) == 0) )
      {
       MagellanXIEId = DeviceList[loop].id;
      };

     XFree( DeviceTypeName );
    };

   if ( MagellanXIEId )
    {
     MagellanXIEDevice = XOpenDevice( display, MagellanXIEId );

     DeviceMotionNotify( MagellanXIEDevice,  
                         MagellanXIEMotionEventType,
                         MagellanXIEMotionEventClass );
     DeviceButtonPress( MagellanXIEDevice,  
                        MagellanXIEButtonPressEventType,
                        MagellanXIEButtonPressEventClass );
     DeviceButtonRelease( MagellanXIEDevice, 
                          MagellanXIEButtonReleaseEventType,
                          MagellanXIEButtonReleaseEventClass );

     MagellanNotifyClass[0] = MagellanXIEMotionEventClass;
     MagellanNotifyClass[1] = MagellanXIEButtonPressEventClass;
     MagellanNotifyClass[2] = MagellanXIEButtonReleaseEventClass;

     MagellanNotifyType[0] = MagellanXIEMotionEventType;
     MagellanNotifyType[1] = MagellanXIEButtonPressEventType;
     MagellanNotifyType[2] = MagellanXIEButtonReleaseEventType;

     XSelectExtensionEvent( display, window, MagellanNotifyClass, 3 );
     XFlush( display );

#ifdef MagellanDEBUG
     printf("Motion=[%d|%d] ButtonPress=[%d|%d] ButtonRelease=[%d|%d]\n",
             MagellanNotifyClass[0], MagellanNotifyType[0],
             MagellanNotifyClass[1], MagellanNotifyType[1],
             MagellanNotifyClass[2], MagellanNotifyType[2] );
#endif
    }
   else
    MagellanExist = FALSE;
   XFreeDeviceList( DeviceList );
  }
 else
  MagellanExist = FALSE;

 return MagellanExist; 
}


int MagellanXIEClose(display)
Display *display;
{
 if ( MagellanXIEDevice != NULL )
  XCloseDevice( display, MagellanXIEDevice ); 
 MagellanXIEDevice = NULL;
 
 return TRUE;
}

int MagellanXIETranslateEvent( display, Event, MagellanEvent,
			       MagellanTranslationScale, MagellanRotationScale )
Display *display;
XEvent *Event;
MagellanFloatEvent *MagellanEvent;
double MagellanTranslationScale;
double MagellanRotationScale;
{
XDeviceMotionEvent *MotionPtr;
XDeviceButtonEvent *ButtonPtr;

 MagellanEvent->MagellanType = FALSE;
 if ( MagellanXIEDevice == NULL )
  return FALSE;

 if ( Event->type == MagellanNotifyType[0] )
  {
   MotionPtr = (XDeviceMotionEvent *) Event;
#ifdef MagellanDEBUG
 printf("Event: type = %d  first = %d axes = %d \n", 
         Event->type, MotionPtr->first_axis, MotionPtr->axes_count );
#endif
   if ( (MotionPtr->axes_count == 6) &&
        (MotionPtr->first_axis == 0) )
    {
     MagellanEvent->MagellanType = MagellanInputMotionEvent;
     MagellanEvent->MagellanData[ MagellanX ] =
      (double) MotionPtr->axis_data[0] * MagellanTranslationScale;
     MagellanEvent->MagellanData[ MagellanY ] =
      (double) MotionPtr->axis_data[1] * MagellanTranslationScale;
     MagellanEvent->MagellanData[ MagellanZ ] =
      (double) MotionPtr->axis_data[2] * MagellanTranslationScale;
     MagellanEvent->MagellanData[ MagellanA ] =
      (double) MotionPtr->axis_data[3] * MagellanRotationScale;
     MagellanEvent->MagellanData[ MagellanB ] =
      (double) MotionPtr->axis_data[4] * MagellanRotationScale;
     MagellanEvent->MagellanData[ MagellanC ] =
      (double) MotionPtr->axis_data[5] * MagellanRotationScale;
     MagellanEvent->MagellanPeriod = 60;
     return MagellanInputMotionEvent;
    }
  };

 if ( Event->type == MagellanNotifyType[1] )
  {
   ButtonPtr = (XDeviceButtonEvent *) Event;
#ifdef MagellanDEBUG
 printf("Event: type = %d  button = %d \n", 
         Event->type, ButtonPtr->button );
#endif
   MagellanEvent->MagellanType = MagellanInputButtonPressEvent;
   MagellanEvent->MagellanButton = ButtonPtr->button;
   return MagellanInputButtonPressEvent;
  }

 if ( Event->type == MagellanNotifyType[2] )
  {
   ButtonPtr = (XDeviceButtonEvent *) Event;
#ifdef MagellanDEBUG
 printf("Event: type = %d  button = %d \n", 
         Event->type, ButtonPtr->button );
#endif
   MagellanEvent->MagellanType = MagellanInputButtonReleaseEvent;
   MagellanEvent->MagellanButton = ButtonPtr->button;
   return MagellanInputButtonReleaseEvent;
  };

 return FALSE;
}

Bool MagellanXIECheckMotionEvent( display, event, arg )
Display *display;
XEvent *event;
char *arg;
{
 if ( event->type == MagellanNotifyType[0] )
  return True;
 return False;
}

int MagellanXIERemoveMotionEvents( display )
Display *display;
{
XEvent event;

 while( XCheckIfEvent(display,&event,MagellanXIECheckMotionEvent,NULL) == True )
  ;

 return TRUE;
}

